import os.path
from tempfile import TemporaryDirectory

from ocdskit.cli.__main__ import main
from tests import assert_command, assert_command_error, read

schema = read('test-schema.json')

schema_with_enum = '''{
  "properties": {
    "closedStringNull": {
      "type": [
        "string",
        "null"
      ],
      "codelist": "a.csv",
      "openCodelist": false,
      "enum": [
        "foo",
        "bar",
        null
      ]
    },
    "closedArrayNull": {
      "type": [
        "array",
        "null"
      ],
      "codelist": "b.csv",
      "openCodelist": false,
      "items": {
        "type": "string",
        "enum": [
          "foo",
          "bar"
        ]
      }
    },
    "closedString": {
      "type": "string",
      "codelist": "c.csv",
      "openCodelist": false,
      "enum": [
        "foo",
        "bar"
      ]
    },
    "closedDisorder": {
      "type": "string",
      "codelist": "d.csv",
      "openCodelist": false,
      "enum": [
        "bar",
        "foo"
      ]
    },
    "open": {
      "type": [
        "string",
        "null"
      ],
      "codelist": "a.csv",
      "openCodelist": true
    },
    "same": {
      "type": [
        "string",
        "null"
      ],
      "codelist": "a.csv",
      "openCodelist": true
    }
  }
}
'''

schema_with_modification = '''{
  "properties": {
    "closedStringNull": {
      "type": [
        "string",
        "null"
      ],
      "codelist": "a.csv",
      "openCodelist": false,
      "enum": [
        "foo",
        "bar",
        "baz",
        null
      ]
    },
    "closedArrayNull": {
      "type": [
        "array",
        "null"
      ],
      "codelist": "b.csv",
      "openCodelist": false,
      "items": {
        "type": "string",
        "enum": [
          "foo"
        ]
      }
    },
    "closedString": {
      "type": "string",
      "codelist": "c.csv",
      "openCodelist": false,
      "enum": [
        "foo",
        "bar"
      ]
    },
    "closedDisorder": {
      "type": "string",
      "codelist": "d.csv",
      "openCodelist": false,
      "enum": [
        "bar",
        "foo"
      ]
    },
    "open": {
      "type": [
        "string",
        "null"
      ],
      "codelist": "a.csv",
      "openCodelist": true
    },
    "same": {
      "type": [
        "string",
        "null"
      ],
      "codelist": "a.csv",
      "openCodelist": true
    }
  }
}
'''

codelist = 'Code\nfoo\nbar\n'


def test_command(monkeypatch, tmpdir):
    tmpdir.join('release-schema.json').write(schema)

    tmpdir.mkdir('codelists')
    for basename in ('a', 'b', 'c', 'd'):
        tmpdir.join('codelists', '{}.csv'.format(basename)).write(codelist)

    assert_command(monkeypatch, main, ['set-closed-codelist-enums', str(tmpdir)], '')
    assert tmpdir.join('release-schema.json').read() == schema_with_enum


def test_unused_codelists(monkeypatch, caplog, tmpdir):
    tmpdir.join('release-schema.json').write(schema)

    tmpdir.mkdir('codelists')
    for basename in ('a', 'b', 'c', 'd', 'e'):
        tmpdir.join('codelists', '{}.csv'.format(basename)).write(codelist)

    assert_command(monkeypatch, main, ['set-closed-codelist-enums', str(tmpdir)], '')
    assert tmpdir.join('release-schema.json').read() == schema_with_enum

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert caplog.records[0].message == 'unused codelists: e.csv'


def test_missing_codelists(monkeypatch, caplog, tmpdir):
    tmpdir.join('release-schema.json').write(schema)

    tmpdir.mkdir('codelists')

    excinfo = assert_command_error(monkeypatch, main, ['set-closed-codelist-enums', str(tmpdir)], error=KeyError)

    assert tmpdir.join('release-schema.json').read() == schema

    assert len(caplog.records) == 0
    assert excinfo.value.args == ('a.csv',)


def test_missing_targets(monkeypatch, caplog, tmpdir):
    tmpdir.join('release-schema.json').write(schema)

    tmpdir.mkdir('codelists')
    for basename in ('a', 'b', 'c', 'd', '+e'):
        tmpdir.join('codelists', '{}.csv'.format(basename)).write(codelist)

    excinfo = assert_command_error(monkeypatch, main, ['set-closed-codelist-enums', str(tmpdir)], error=KeyError)

    assert tmpdir.join('release-schema.json').read() == schema

    assert len(caplog.records) == 0
    assert excinfo.value.args == ('e.csv',)


def test_conflicting_codelists(monkeypatch, caplog, tmpdir):
    with TemporaryDirectory() as d:
        for directory in (tmpdir, d):
            os.mkdir(os.path.join(directory, 'codelists'))

            with open(os.path.join(directory, 'release-schema.json'), 'w') as f:
                f.write(schema)

        for basename in ('a', 'b', 'c', 'd'):
            tmpdir.join('codelists', '{}.csv'.format(basename)).write(codelist)

        with open(os.path.join(d, 'codelists', 'a.csv'), 'w') as f:
            f.write('Code\nbaz\n')

        assert_command(monkeypatch, main, ['set-closed-codelist-enums', str(tmpdir), d], '')
        with open(os.path.join(d, 'release-schema.json')) as f:
            assert f.read() == schema_with_enum

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == 'ERROR'
        assert caplog.records[0].message == 'conflicting codelists: a.csv'


def test_modified_codelists(monkeypatch, tmpdir):
    with TemporaryDirectory() as d:
        for directory in (tmpdir, d):
            os.mkdir(os.path.join(directory, 'codelists'))

            with open(os.path.join(directory, 'release-schema.json'), 'w') as f:
                f.write(schema)

        for basename in ('a', 'b', 'c', 'd'):
            tmpdir.join('codelists', '{}.csv'.format(basename)).write(codelist)

        with open(os.path.join(d, 'codelists', '+a.csv'), 'w') as f:
            f.write('Code\nbaz\n')

        with open(os.path.join(d, 'codelists', '-b.csv'), 'w') as f:
            f.write('Code,Description\nbar,bzz\n')

        assert_command(monkeypatch, main, ['set-closed-codelist-enums', str(tmpdir), d], '')
        with open(os.path.join(d, 'release-schema.json')) as f:
            assert f.read() == schema_with_modification



import os.path
import sys
from difflib import ndiff
from io import BytesIO, StringIO, TextIOWrapper
from itertools import zip_longest
from unittest.mock import patch

import pytest

import ocdskit.util


def path(filename):
    return os.path.join('tests', 'fixtures', filename)


def read(filename, mode='rt', encoding=None, **kwargs):
    with open(path(filename), mode, encoding=encoding, **kwargs) as f:
        return f.read()


def assert_equal(actual, expected, ordered=True):
    if ordered:
        assert actual == expected, ''.join(ndiff(actual.splitlines(1), expected.splitlines(1)))
    else:
        for a, b in zip_longest(actual.split('\n'), expected.split('\n'), fillvalue='{}'):
            if a != b != '':
                assert ocdskit.util.jsonlib.loads(a) == ocdskit.util.jsonlib.loads(b), '\n{}\n{}'.format(a, b)


@patch('sys.stdout', new_callable=StringIO)
def run_command(monkeypatch, main, args, stdout):
    monkeypatch.setattr(sys, 'argv', ['ocdskit'] + args)
    main()

    return stdout.getvalue()


# Similar to `run_command`, but with `pytest.raises` block.
@patch('sys.stdout', new_callable=StringIO)
def assert_command_error(monkeypatch, main, args, stdout, expected='', error=SystemExit):
    with pytest.raises(error) as excinfo:
        monkeypatch.setattr(sys, 'argv', ['ocdskit'] + args)
        main()

    actual = stdout.getvalue()

    assert actual == expected, '\n{}\n{}'.format(actual, expected)
    if error is SystemExit:
        assert excinfo.value.code == 1

    return excinfo


def assert_command(monkeypatch, main, args, expected, ordered=True):
    actual = run_command(monkeypatch, main, args)

    if os.path.isfile(path(expected)):
        expected = read(expected, newline='')

    assert_equal(actual, expected, ordered=ordered)


@patch('sys.stdout', new_callable=StringIO)
def run_streaming(monkeypatch, main, args, stdin, stdout):
    if not isinstance(stdin, bytes):
        stdin = b''.join(read(filename, 'rb') for filename in stdin)

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))):
        monkeypatch.setattr(sys, 'argv', ['ocdskit'] + args)
        main()

    return stdout.getvalue()


# Similar to `run_streaming`, but with `pytest.raises` block.
@patch('sys.stdout', new_callable=StringIO)
def assert_streaming_error(monkeypatch, main, args, stdin, stdout, expected='', error=SystemExit):
    if not isinstance(stdin, bytes):
        stdin = b''.join(read(filename, 'rb') for filename in stdin)

    with pytest.raises(error) as excinfo:
        with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))):
            monkeypatch.setattr(sys, 'argv', ['ocdskit'] + args)
            main()

    actual = stdout.getvalue()

    assert actual == expected, '\n{}\n{}'.format(actual, expected)
    if error is SystemExit:
        assert excinfo.value.code == 1

    return excinfo


def assert_streaming(monkeypatch, main, args, stdin, expected, ordered=True):
    actual = run_streaming(monkeypatch, main, args, stdin)

    if not isinstance(expected, str):
        expected = ''.join(read(filename) for filename in expected)

    assert_equal(actual, expected, ordered=ordered)



from ocdskit.cli.__main__ import main
from tests import assert_streaming


def test_command(monkeypatch):
    assert_streaming(monkeypatch, main, ['split-record-packages', '1'], ['realdata/record-package_package.json'],
                     ['realdata/record-package_split.json'])



import json

import pytest
from ocdsmerge import Merger
from ocdsmerge.util import get_release_schema_url, get_tags

from ocdskit.packager import Packager
from tests import read


@pytest.mark.vcr()
def test_output_package_no_streaming():
    data = [json.loads(read(filename)) for filename in
            ('realdata/release-package-1.json', 'realdata/release-package-2.json')]

    with Packager() as packager:
        packager.package['version'] = '1.1'
        packager.add(data)

        prefix = packager.version.replace('.', '__') + '__'
        tag = next(tag for tag in reversed(get_tags()) if tag.startswith(prefix))
        schema = get_release_schema_url(tag)

        actual = next(packager.output_package(Merger(schema)))

    assert actual == json.loads(read('realdata/record-package_package.json'))



import itertools
import json
from decimal import Decimal

import ijson

from ocdskit.exceptions import UnknownFormatError

try:
    import orjson

    jsonlib = orjson
    USING_ORJSON = True
except ImportError:
    jsonlib = json
    USING_ORJSON = False


# See `grouper` recipe: https://docs.python.org/3.8/library/itertools.html#recipes
def grouper(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


# https://stackoverflow.com/questions/21663800/python-make-a-list-generator-json-serializable/46841935#46841935
class SerializableGenerator(list):
    def __init__(self, iterable):
        iterator = iter(iterable)
        try:
            # If `iter()` is omitted, then `__iter__` won't exhaust `head`.
            self.head = iter([next(iterator)])
            # Adding an item to the list ensures `__bool__` and `__len__` work.
            self.append(iterator)
        except StopIteration:
            # `__iter__` requires `head` to be set.
            self.head = []

    def __iter__(self):
        # `*self[:1]` is used, because `self[0]` raises IndexError when `iterable` is empty.
        return itertools.chain(self.head, *self[:1])


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Decimal):
            return float(obj)
        # https://docs.python.org/3/library/json.html#json.JSONEncoder.default
        try:
            iter(obj)
        except TypeError:
            pass
        else:
            return SerializableGenerator(obj)
        return json.JSONEncoder.default(self, obj)


def iterencode(data, ensure_ascii=False, **kwargs):
    """
    Returns a generator that yields each string representation as available.
    """
    if 'indent' not in kwargs:
        kwargs['separators'] = (',', ':')
    return JSONEncoder(ensure_ascii=ensure_ascii, **kwargs).iterencode(data)


def json_dump(data, io, ensure_ascii=False, **kwargs):
    """
    Dumps JSON to a file-like object.
    """
    if 'indent' not in kwargs:
        kwargs['separators'] = (',', ':')
    json.dump(data, io, ensure_ascii=ensure_ascii, cls=JSONEncoder, **kwargs)


def json_dumps(data, ensure_ascii=False, indent=None, sort_keys=False, **kwargs):
    """
    Dumps JSON to a string, and returns it.
    """
    # orjson doesn't support `ensure_ascii` if `True`, `indent` if not `2` or other arguments except for `sort_keys`.
    if not USING_ORJSON or ensure_ascii or indent and indent != 2 or kwargs:
        if not indent:
            kwargs['separators'] = (',', ':')
        return json.dumps(data, cls=JSONEncoder, ensure_ascii=ensure_ascii, indent=indent, sort_keys=sort_keys,
                          **kwargs)

    option = 0
    if indent:
        option |= orjson.OPT_INDENT_2
    if sort_keys:
        option |= orjson.OPT_SORT_KEYS

    # orjson dumps to bytes.
    return orjson.dumps(data, default=JSONEncoder().default, option=option).decode()


def get_ocds_minor_version(data):
    """
    Returns the OCDS minor version of the record package, release package, record or release.
    """
    if is_package(data):
        if 'version' in data:
            return data['version']
        return '1.0'
    if is_record(data):
        if any('parties' in release for release in data['releases']):
            return '1.1'
        return '1.0'
    # release
    if 'parties' in data:
        return '1.1'
    return '1.0'


def is_package(data):
    """
    Returns whether the data is a record package or release package.
    """
    return is_record_package(data) or is_release_package(data)


def is_record_package(data):
    """
    Returns whether the data is a record package.

    A record package has a required ``records`` field. Its other required fields are shared with release packages.
    """
    return 'records' in data


def is_record(data):
    """
    Returns whether the data is a record.

    A record has required ``releases`` and ``ocid`` fields.
    """
    return 'releases' in data and 'ocid' in data


def is_release_package(data):
    """
    Returns whether the data is a release package.

    A release package has a required ``releases`` field. Its other required fields are shared with record packages.
    To distinguish a release package from a record, we test for the absence of the ``ocid`` field.
    """
    return 'releases' in data and 'ocid' not in data


def is_release(data):
    """
    Returns whether the data is a release (embedded or linked, individual or compiled).
    """
    return 'date' in data


def is_compiled_release(data):
    """
    Returns whether the data is a compiled release (embedded or linked).
    """
    return 'tag' in data and isinstance(data['tag'], list) and 'compiled' in data['tag']


def is_linked_release(data):
    """
    Returns whether the data is a linked release.

    A linked release has required ``url`` and ``date`` fields and an optional ``tag`` field. An embedded release has
    required ``date`` and ``tag`` fields (among others), and it can have a ``url`` field as an additional field.

    To distinguish a linked release from an embedded release, we test for the presence of the required ``url`` field
    and test whether the number of fields is fewer than three.
    """
    return 'url' in data and len(data) <= 3


def _empty_package(uri, publisher, published_date, version):
    if publisher is None:
        publisher = {}

    return {
        'uri': uri,
        'publisher': publisher,
        'publishedDate': published_date,
        'license': None,
        'publicationPolicy': None,
        'version': version,
        'extensions': {},
    }


def detect_format(path, root_path=''):
    """
    Returns the format of OCDS data, and whether the OCDS data is concatenated or in an array.

    If the OCDS data is concatenated or in an array, assumes that all items have the same format as the first item.

    :param str path: the path to a file
    :param str root_path: the path to the OCDS data within the file
    :returns: the format, whether data is concatenated, and whether data is in an array
    :rtype: tuple
    :raises UnknownFormatError: if the format cannot be detected
    """
    with open(path, 'rb') as f:
        events = iter(ijson.parse(f, multiple_values=True))

        while True:
            prefix, event, value = next(events)
            if prefix == root_path:
                break

        if prefix:
            prefix += '.'

        if event == 'start_array':
            prefix += 'item.'
        elif event != 'start_map':
            raise UnknownFormatError('top-level JSON value is a {}'.format(event))

        records_prefix = '{}records'.format(prefix)
        releases_prefix = '{}releases'.format(prefix)
        ocid_prefix = '{}ocid'.format(prefix)
        tag_item_prefix = '{}tag.item'.format(prefix)

        has_records = False
        has_releases = False
        has_ocid = False
        has_tag = False
        is_compiled = False
        is_array = event == 'start_array'

        for prefix, event, value in events:
            if prefix == records_prefix:
                has_records = True
            elif prefix == releases_prefix:
                has_releases = True
            elif prefix == ocid_prefix:
                has_ocid = True
            elif prefix == tag_item_prefix:
                has_tag = True
                if value == 'compiled':
                    is_compiled = True
            if not prefix and event not in ('end_array', 'end_map', 'map_key'):
                return _detect_format_result(True, is_array, has_records, has_releases, has_ocid, has_tag, is_compiled)

        return _detect_format_result(False, is_array, has_records, has_releases, has_ocid, has_tag, is_compiled)


def _detect_format_result(is_concatenated, is_array, has_records, has_releases, has_ocid, has_tag, is_compiled):
    if has_records:
        detected_format = 'record package'
    elif has_releases and has_ocid:
        detected_format = 'record'
    elif has_releases:
        detected_format = 'release package'
    elif is_compiled:
        detected_format = 'compiled release'
    elif has_tag:
        detected_format = 'release'
    elif has_ocid:
        detected_format = 'versioned release'
    else:
        if is_array:
            infix = 'array'
        else:
            infix = 'object'
        raise UnknownFormatError('top-level JSON value is a non-OCDS {}'.format(infix))

    return (detected_format, is_concatenated, is_array)


def _empty_record_package(uri='', publisher=None, published_date='', version=None):
    package = _empty_package(uri, publisher, published_date, version)
    package['packages'] = []
    package['records'] = []
    return package


def _empty_release_package(uri='', publisher=None, published_date='', version=None):
    package = _empty_package(uri, publisher, published_date, version)
    package['releases'] = []
    return package


def _update_package_metadata(output, package):
    for field in ('publisher', 'license', 'publicationPolicy'):
        if field in package:
            output[field] = package[field]

    # We use an insertion-ordered dict to keep extensions in order without duplication.
    if 'extensions' in package:
        output['extensions'].update(dict.fromkeys(package['extensions']))


def _resolve_metadata(output, field):
    if output[field]:
        output[field] = list(output[field])
    else:
        del output[field]


def _remove_empty_optional_metadata(output):
    for field in ('license', 'publicationPolicy', 'version'):
        if output[field] is None:
            del output[field]


def _cast_as_list(value):
    if isinstance(value, str):
        return [value]
    return sorted(value)



from ocdskit.cli.__main__ import main
from tests import assert_command

content = b'{"lorem":"ipsum"}'
invalid = b'{"lorem":"ipsum"'


def test_command(monkeypatch, tmpdir):
    p = tmpdir.join('test.json')
    p.write(content)

    assert_command(monkeypatch, main, ['indent', str(p)], '')

    assert p.read() == '{\n  "lorem": "ipsum"\n}\n'


def test_indent(monkeypatch, tmpdir):
    p = tmpdir.join('test.json')
    p.write(content)

    assert_command(monkeypatch, main, ['indent', '--indent', '4', str(p)], '')

    assert p.read() == '{\n    "lorem": "ipsum"\n}\n'


def test_ascii(monkeypatch, tmpdir):
    p = tmpdir.join('test.json')
    p.write(b'{"lorem":"ips\\u00fam"}')

    assert_command(monkeypatch, main, ['--ascii', 'indent', str(p)], '')

    assert p.read() == '{\n  "lorem": "ips\\u00fam"\n}\n'


def test_command_recursive(monkeypatch, tmpdir):
    tmpdir.join('test.json').write(content)
    tmpdir.join('test.txt').write(content)

    assert_command(monkeypatch, main, ['indent', '--recursive', str(tmpdir)], '')

    assert tmpdir.join('test.json').read() == '{\n  "lorem": "ipsum"\n}\n'
    assert tmpdir.join('test.txt').read() == content.decode()


def test_command_directory(monkeypatch, caplog, tmpdir):
    assert_command(monkeypatch, main, ['indent', str(tmpdir)], '')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert caplog.records[0].message.endswith(' is a directory. Set --recursive to recurse into directories.')


def test_command_nonexistent(monkeypatch, caplog):
    assert_command(monkeypatch, main, ['indent', 'nonexistent'], '')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert caplog.records[0].message == 'nonexistent: No such file or directory'


def test_command_invalid_json(monkeypatch, caplog, tmpdir):
    p = tmpdir.join('test.json')
    p.write(invalid)

    assert_command(monkeypatch, main, ['indent', str(p)], '')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert ' is not valid JSON. (json.decoder.JSONDecodeError: ' in caplog.records[0].message



import logging
import os.path

from ocdskit.cli.commands.base import OCDSCommand
from ocdskit.exceptions import UnknownFormatError
from ocdskit.util import detect_format

logger = logging.getLogger('ocdskit')


class Command(OCDSCommand):
    name = 'detect-format'
    help = 'reads OCDS files, and reports whether each is a release, record, package, etc.'

    def add_arguments(self):
        self.add_argument('file', help='OCDS files', nargs='+')
        self.add_argument('-r', '--recursive', help='recursively indent JSON files', action='store_true')

    def handle(self):
        for file in self.args.file:
            if os.path.isfile(file):
                self.detect_format(file)
            elif self.args.recursive:
                for root, _, files in os.walk(file):
                    for name in files:
                        if not name.startswith('.'):
                            self.detect_format(os.path.join(root, name))
            elif os.path.isdir(file):
                logger.warning('%s is a directory. Set --recursive to recurse into directories.', file)
            else:
                logger.error('%s: No such file or directory', file)

    def detect_format(self, path):
        try:
            _print(path, *detect_format(path, self.args.root_path))
        except UnknownFormatError as e:
            logger.warning('%s: unknown (%s)', path, e)


def _print(path, detected_format, is_concatenated, is_array):
    string = detected_format

    if is_array:
        string = 'a JSON array of {}s'.format(string)

    if is_concatenated:
        string = 'concatenated JSON, starting with {}'.format(string)

    print('{}: {}'.format(path, string))



class OCDSKitError(Exception):
    """Base class for exceptions from within this package"""


class CommandError(OCDSKitError):
    """Errors from within this package's CLI"""


class InconsistentVersionError(OCDSKitError):
    """Raised if the versions are inconsistent across packages to compile"""

    def __init__(self, message, earlier_version=None, current_version=None):
        self.earlier_version = earlier_version
        self.current_version = current_version
        super().__init__(message)


class MissingColumnError(OCDSKitError):
    """Raised if the column by which to order is missing"""


class UnknownFormatError(OCDSKitError):
    """Raised if the format of a file can't be determined"""


class MissingOcidKeyError(OCDSKitError, KeyError):
    """Raised if a release to be merged is missing an ``ocid`` field"""


class OCDSKitWarning(UserWarning):
    """Base class for warnings from within this package"""


class MissingRecordsWarning(OCDSKitWarning):
    """Used when the "records" field is missing from a record package when combining packages"""

    def __str__(self):
        return 'item {0} has no "records" field (check that it is a record package)'.format(*self.args)


class MissingReleasesWarning(OCDSKitWarning):
    """Used when the "releases" field is missing from a release package when combining packages"""

    def __str__(self):
        return 'item {0} has no "releases" field (check that it is a release package)'.format(*self.args)



import json

import pytest

from ocdskit.util import (detect_format, get_ocds_minor_version, is_compiled_release, is_linked_release, is_package,
                          is_record, is_record_package, is_release, is_release_package, json_dump)
from tests import path, read


# Same fixture files as in test_detect_format.py, except for concatenated JSON files.
@pytest.mark.parametrize('filename,expected', [
    ('record-package_minimal.json', True),
    ('release-package_minimal.json', True),
    ('record_minimal.json', False),
    ('release_minimal.json', False),
    ('realdata/compiled-release-1.json', False),
    ('realdata/versioned-release-1.json', False),
    ('release-packages.json', False),
    ('detect-format_whitespace.json', False),
])
def test_is_package(filename, expected):
    assert is_package(json.loads(read(filename))) == expected


@pytest.mark.parametrize('filename,expected', [
    ('record-package_minimal.json', True),
    ('release-package_minimal.json', False),
    ('record_minimal.json', False),
    ('release_minimal.json', False),
    ('realdata/compiled-release-1.json', False),
    ('realdata/versioned-release-1.json', False),
    ('release-packages.json', False),
    ('detect-format_whitespace.json', False),
])
def test_is_record_package(filename, expected):
    assert is_record_package(json.loads(read(filename))) == expected


@pytest.mark.parametrize('filename,expected', [
    ('record-package_minimal.json', False),
    ('release-package_minimal.json', False),
    ('record_minimal.json', True),
    ('release_minimal.json', False),
    ('realdata/compiled-release-1.json', False),
    ('realdata/versioned-release-1.json', False),
    ('release-packages.json', False),
    ('detect-format_whitespace.json', False),
])
def test_is_record(filename, expected):
    assert is_record(json.loads(read(filename))) == expected


@pytest.mark.parametrize('filename,expected', [
    ('record-package_minimal.json', False),
    ('release-package_minimal.json', True),
    ('record_minimal.json', False),
    ('release_minimal.json', False),
    ('realdata/compiled-release-1.json', False),
    ('realdata/versioned-release-1.json', False),
    ('release-packages.json', False),
    ('detect-format_whitespace.json', False),
])
def test_is_release_package(filename, expected):
    assert is_release_package(json.loads(read(filename))) == expected


@pytest.mark.parametrize('filename,expected', [
    ('record-package_minimal.json', False),
    ('release-package_minimal.json', False),
    ('record_minimal.json', False),
    ('release_minimal.json', True),
    ('realdata/compiled-release-1.json', True),
    ('realdata/versioned-release-1.json', False),
    ('release-packages.json', False),
    ('detect-format_whitespace.json', True),
])
def test_is_release(filename, expected):
    assert is_release(json.loads(read(filename))) == expected


@pytest.mark.parametrize('data,expected', [
    ({'url': 'http://example.com', 'date': '2001-02-03T04:05:06Z', 'tag': ['compiled']}, True),
    ({'url': 'http://example.com', 'date': '2001-02-03T04:05:06Z', 'tag': ['tender']}, False),
    ({'url': 'http://example.com', 'date': '2001-02-03T04:05:06Z', 'tag': None}, False),
    ({'url': 'http://example.com', 'date': '2001-02-03T04:05:06Z'}, False),
])
def test_is_compiled_release(data, expected):
    assert is_compiled_release(data) == expected


@pytest.mark.parametrize('data,expected', [
    ({'url': 'http://example.com'}, True),
    ({'url': 'http://example.com', 'date': '2001-02-03T04:05:06Z'}, True),
    ({'url': 'http://example.com', 'date': '2001-02-03T04:05:06Z', 'tag': ['tender']}, True),
    ({'url': 'http://example.com', 'date': '2001-02-03T04:05:06Z', 'tag': ['tender'], 'id': '1'}, False),
])
def test_is_linked_release(data, expected):
    assert is_linked_release(data) == expected


@pytest.mark.parametrize('data,expected', [
    ({'records': [], 'version': '99.99'}, '99.99'),
    ({'records': []}, '1.0'),
    ({'releases': [], 'version': '99.99'}, '99.99'),
    ({'releases': []}, '1.0'),
    # An additional `version` field is provided on records and releases, to be sure it isn't used.
    ({'version': '99.99', 'ocid': 'ocds-213czf-1', 'releases': [{'parties': []}]}, '1.1'),
    ({'version': '99.99', 'ocid': 'ocds-213czf-1', 'releases': []}, '1.0'),
    ({'version': '99.99', 'ocid': 'ocds-213czf-1', 'parties': []}, '1.1'),
    ({'version': '99.99', 'ocid': 'ocds-213czf-1'}, '1.0'),
])
def test_get_ocds_minor_version(data, expected):
    assert get_ocds_minor_version(data) == expected


@pytest.mark.parametrize('data,expected', [
    (iter([]), '[]'),
    ({'a': 1, 'b': 2}, '{"a":1,"b":2}'),
])
def test_json_dump(data, expected, tmpdir):
    p = tmpdir.join('test.json')

    with open(p, 'w') as f:
        json_dump(data, f)

    assert p.read() == expected


@pytest.mark.parametrize('filename,expected', [
    ('record-package_minimal.json', ('record package', False, False)),
    ('release-package_minimal.json', ('release package', False, False)),
    ('record_minimal.json', ('record', False, False)),
    ('release_minimal.json', ('release', False, False)),
    ('realdata/compiled-release-1.json', ('compiled release', False, False)),
    ('realdata/versioned-release-1.json', ('versioned release', False, False)),
    ('release-packages.json', ('release package', False, True)),
    ('release-packages.jsonl', ('release package', True, True)),
    ('detect-format_mixed.json', ('release', True, False)),
    ('detect-format_whitespace.json', ('release', False, False)),
])
def test_detect_format(filename, expected):
    result = detect_format(path(filename))

    assert result == expected



import csv
import itertools
import json
import logging
import os.path
from copy import deepcopy

from ocdskit.cli.commands.base import BaseCommand
from ocdskit.util import _cast_as_list, json_dump

logger = logging.getLogger('ocdskit')

meta_schema_exceptions = {
    'meta-schema.json',
    'meta-schema-patch.json',
}


class Command(BaseCommand):
    name = 'set-closed-codelist-enums'
    help = 'sets the enum in a JSON Schema to match the codes in the CSV files of closed codelists'

    def __init__(self, subparsers):
        super().__init__(subparsers)

        self.codelists = {}
        self.codelists_seen = set()

    def add_arguments(self):
        self.add_argument('standard', help='path to directory containing standard JSON Schema files')
        self.add_argument('extension', help='paths to directories containing extension JSON Schema files', nargs='*')

    def handle(self):
        if self.args.extension:
            self.collect_codelists(self.args.standard)
            self.codelists_seen = set(self.codelists)
            directories = self.args.extension
        else:
            directories = [self.args.standard]

        for directory in directories:
            self.collect_codelists(directory)
            self.update_json_schema(directory)

        codelists_not_seen = []
        for codelist in self.codelists:
            if codelist not in self.codelists_seen:
                codelists_not_seen.append(codelist)

        if codelists_not_seen:
            logger.error('unused codelists: %s', ' '.join(codelists_not_seen))

    def collect_codelists(self, directory):
        for root, _, files in os.walk(directory):
            for name in files:
                if name.endswith('.csv'):
                    with open(os.path.join(root, name)) as f:
                        reader = csv.DictReader(f)
                        row = next(reader)
                        if 'Code' in row:
                            codes = [row['Code'] for row in itertools.chain([row], reader)]
                        else:
                            codes = []

                    if codes:
                        if name.startswith('+'):
                            # KeyError if codelist doesn't exist.
                            self.codelists[name[1:]] += codes
                        elif name.startswith('-'):
                            # KeyError if codelist doesn't exist.
                            self.codelists[name[1:]] = [code for code in self.codelists[name[1:]] if code not in codes]
                        elif name in self.codelists:
                            if self.codelists[name] != codes:
                                logger.error('conflicting codelists: %s', name)
                        else:
                            self.codelists[name] = codes

    # This method is similar to `validate_codelist_enum` in `test_json.py`.
    def update_codelist_enum(self, data):
        if isinstance(data, list):
            return [self.update_codelist_enum(item) for item in data]

        if isinstance(data, dict) and 'codelist' not in data:
            return {key: self.update_codelist_enum(value) for key, value in data.items()}

        if isinstance(data, dict) and 'codelist' in data:
            self.codelists_seen.add(data['codelist'])

            if not data['openCodelist']:
                codes = self.codelists[data['codelist']]

                types = _cast_as_list(data['type'])

                if 'string' in types:
                    if 'null' in types:
                        codes.append(None)
                    if 'enum' not in data or set(data['enum']) != set(codes):
                        data['enum'] = codes
                elif 'array' in types:
                    if 'enum' not in data['items'] or set(data['items']['enum']) != set(codes):
                        data['items']['enum'] = codes

        return data

    def update_json_schema(self, directory):
        for root, _, files in os.walk(directory):
            for name in files:
                if name.endswith('.json') and name not in meta_schema_exceptions:
                    path = os.path.join(root, name)

                    with open(path) as f:
                        data = json.load(f)

                    # If the JSON file is a JSON Schema file.
                    if any(field in data for field in ('$schema', 'definitions', 'properties')):
                        expected = deepcopy(data)
                        self.update_codelist_enum(data)

                        if expected != data:
                            with open(path, 'w') as f:
                                json_dump(data, f, indent=2)
                                f.write('\n')



import json

import pytest
from jsonpointer import set_pointer

from ocdskit.cli.__main__ import main
from tests import assert_streaming, assert_streaming_error, read, run_streaming


def test_command_record_package(monkeypatch, caplog):
    assert_streaming(monkeypatch, main, ['upgrade', '1.0:1.1'],
                     ['realdata/record-package_1.0.json'],
                     ['realdata/record-package_1.1.json'], ordered=False)

    assert len(caplog.records) == 0


def test_command_record(monkeypatch, caplog):
    assert_streaming(monkeypatch, main, ['upgrade', '1.0:1.1', '--root-path', 'records.item'],
                     ['realdata/record-package_package.json'],
                     ['realdata/record-package_package_1.1.json'], ordered=False)

    assert len(caplog.records) == 0


def test_command_release_package_buyer_procuring_entity_suppliers(monkeypatch, caplog):
    assert_streaming(monkeypatch, main, ['upgrade', '1.0:1.1'],
                     ['realdata/release-package_1.0-1.json'],
                     ['realdata/release-package_1.1-1.json'], ordered=False)

    assert len(caplog.records) == 0


def test_command_release_package_transactions(monkeypatch, caplog):
    assert_streaming(monkeypatch, main, ['upgrade', '1.0:1.1'],
                     ['realdata/release-package_1.0-2.json'],
                     ['realdata/release-package_1.1-2.json'], ordered=False)

    assert len(caplog.records) == 0


def test_command_release_tenderers_amendment(monkeypatch, caplog):
    assert_streaming(monkeypatch, main, ['upgrade', '1.0:1.1'],
                     ['release_1.0.json'],
                     ['release_1.1.json'], ordered=False)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert caplog.records[0].message == (
        'party in "supplier" role differs from party in ["tenderer"] roles:\n'
        '{"name": "Acme Inc.", "identifier": {"id": 1}, "additionalIdentifiers": [{"id": "a"}], "id": "3c9756cf8983b14066a034079aa7aae4"}\n'  # noqa: E501
        '{"id": "3c9756cf8983b14066a034079aa7aae4", "name": "Acme Inc.", "identifier": {"id": 1}}'
    )


@pytest.mark.parametrize('pointer', ('parties', 'buyer', 'tender', 'tender/procuringEntity', 'tender/tenderers',
                                     'awards', 'awards/0/suppliers', 'contracts', 'contracts/0/implementation',
                                     'contracts/0/implementation/transactions'))
def test_command_release_field_is_null(pointer, monkeypatch, caplog):
    data = json.loads(read('release_minimal.json'))

    parts = pointer.split('/')
    for i, part in enumerate(parts, 1):
        if i < len(parts):
            if parts[i] == '0':
                value = [{}]
            else:
                value = {}
        else:
            value = None
        set_pointer(data, '/' + '/'.join(parts[:i]), value)

    stdin = json.dumps(data).encode('utf-8')

    # Should not raise an error.
    run_streaming(monkeypatch, main, ['upgrade', '1.0:1.1'], stdin)

    assert len(caplog.records) == 0


def test_command_release_party_id_missing(monkeypatch, caplog):
    data = json.loads(read('release-package_minimal.json'))

    data['releases'][0]['parties'] = [{'name': 'Acme Inc.'}]
    del data['version']

    stdin = json.dumps(data).encode('utf-8')

    # Should not raise an error.
    run_streaming(monkeypatch, main, ['upgrade', '1.0:1.1'], stdin)

    assert len(caplog.records) == 0


def test_command_release_party_roles_missing(monkeypatch, caplog):
    data = json.loads(read('release-package_minimal.json'))

    data['releases'][0]['parties'] = [{'id': '1', 'name': 'Acme Inc.'}]
    data['releases'][0]['buyer'] = {'id': '1'}
    del data['version']

    stdin = json.dumps(data).encode('utf-8')

    # Should not raise an error.
    run_streaming(monkeypatch, main, ['upgrade', '1.0:1.1'], stdin)

    assert len(caplog.records) == 0


def test_command_release_party_roles_str(monkeypatch, caplog):
    data = json.loads(read('release-package_minimal.json'))

    data['releases'][0]['parties'] = [{'id': '1', 'roles': 'buyer'}]
    data['releases'][0]['tender'] = {'procuringEntity': {'id': '1'}}
    del data['version']

    stdin = json.dumps(data).encode('utf-8')

    # Should not raise an error.
    run_streaming(monkeypatch, main, ['upgrade', '1.0:1.1'], stdin)

    assert len(caplog.records) == 0


def test_command_passthrough_package(monkeypatch, caplog):
    assert_streaming(monkeypatch, main, ['upgrade', '1.0:1.1'],
                     ['realdata/record-package_1.1.json'],
                     ['realdata/record-package_1.1.json'])

    assert len(caplog.records) == 0


def test_command_passthrough_release(monkeypatch, caplog):
    assert_streaming(monkeypatch, main, ['upgrade', '1.0:1.1'],
                     ['release_1.1.json'],
                     ['release_1.1.json'])

    assert len(caplog.records) == 0


@pytest.mark.parametrize('versions', ['1.0:1.0', '1.1:1.1'])
def test_command_identity(versions, monkeypatch, caplog):
    assert_streaming(monkeypatch, main, ['upgrade', versions], b'{}', '{}\n')

    assert len(caplog.records) == 0


def test_command_downgrade(monkeypatch, caplog):
    stdin = b'{}'

    assert_streaming_error(monkeypatch, main, ['upgrade', '1.1:1.0'], stdin)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'CRITICAL'
    assert caplog.records[0].message == 'downgrade from 1.1 to 1.0 is not supported'



from ocdskit.cli.commands.base import OCDSCommand
from ocdskit.combine import package_releases
from ocdskit.util import grouper


class Command(OCDSCommand):
    name = 'package-releases'
    help = 'reads releases from standard input, and prints one release package'

    def add_arguments(self):
        self.add_argument('extension', help='add this extension to the package', nargs='*')
        self.add_argument('--size', type=int, help='the maximum number of releases per package')

        self.add_package_arguments('release')

    def handle(self):
        kwargs = self.parse_package_arguments()
        kwargs['extensions'] = self.args.extension

        if self.args.size:  # assume `--size` is reasonable
            for data in grouper(self.items(), self.args.size):
                output = package_releases(list(filter(None, data)), **kwargs)
                self.print(output)
        else:
            output = package_releases(self.items(), **kwargs)
            self.print(output, streaming=True)



import json
import sys
from io import BytesIO, StringIO, TextIOWrapper
from unittest.mock import patch

from ocdskit.cli.__main__ import main
from tests import assert_streaming, read, run_streaming


def test_command(monkeypatch):
    assert_streaming(monkeypatch, main, ['combine-record-packages'],
                     ['record-package_minimal.json', 'record-package_maximal.json', 'record-package_extensions.json'],
                     ['combine-record-packages_minimal-maximal-extensions.json'])


def test_command_no_extensions(monkeypatch):
    assert_streaming(monkeypatch, main, ['combine-record-packages'],
                     ['record-package_minimal.json'],
                     ['combine-record-packages_minimal.json'])


def test_command_uri_published_date(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['combine-record-packages', '--uri', 'http://example.com/x.json',
                                               '--published-date', '2010-01-01T00:00:00Z', '--version', '1.2'],
                           ['record-package_minimal.json'])

    package = json.loads(actual)
    assert package['uri'] == 'http://example.com/x.json'
    assert package['publishedDate'] == '2010-01-01T00:00:00Z'
    assert package['version'] == '1.2'


def test_command_publisher(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['combine-record-packages', '--publisher-name', 'Acme Inc.'],
                           ['record-package_minimal.json'])

    package = json.loads(actual)
    assert package['publisher']['name'] == 'Acme Inc.'


@patch('sys.stderr', new_callable=StringIO)
@patch('sys.stdout', new_callable=StringIO)
def test_command_missing_field(stdout, stderr, monkeypatch):
    stdin = read('release-package_minimal.json', 'rb')

    with patch('sys.stdin', TextIOWrapper(BytesIO(stdin))):
        monkeypatch.setattr(sys, 'argv', ['ocdskit', 'combine-record-packages'])
        main()

    assert stdout.getvalue() == '{"uri":"","publisher":{"name":"Acme"},"publishedDate":"","version":"1.1","records":[]}\n'  # noqa: E501
    assert stderr.getvalue() == 'item 0 has no "records" field (check that it is a record package)\n'


def test_command_packages(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['combine-record-packages'],
                           ['realdata/record-package_package.json', 'realdata/record-package_package.json'])

    package = json.loads(actual)
    assert package['packages'] == [
        'http://www.contratosabiertos.cdmx.gob.mx/api/contrato/OCDS-87SD3T-AD-SF-DRM-063-2015',
        'http://www.contratosabiertos.cdmx.gob.mx/api/contrato/OCDS-87SD3T-AD-SF-DRM-065-2015',
    ]



import json
from io import StringIO
from unittest.mock import patch

from ocdskit.cli.__main__ import main
from tests import path, run_command

expected = '''{
  "required": [
    "array",
    "minItemsArray",
    "object",
    "minPropertiesObject",
    "string",
    "minLengthString",
    "enumString",
    "formatString",
    "patternString"
  ],
  "properties": {
    "optionalArray": {
      "type": "array",
      "minItems": 1,
      "uniqueItems": true
    },
    "array": {
      "type": "array",
      "minItems": 1,
      "uniqueItems": true
    },
    "minItemsArray": {
      "type": "array",
      "minItems": 2,
      "uniqueItems": true
    },
    "optionalObject": {
      "type": "object",
      "minProperties": 1
    },
    "object": {
      "type": "object",
      "minProperties": 1
    },
    "minPropertiesObject": {
      "type": "object",
      "minProperties": 2
    },
    "optionalString": {
      "type": "string",
      "minLength": 1
    },
    "string": {
      "type": "string",
      "minLength": 1
    },
    "minLengthString": {
      "type": "string",
      "minLength": 2
    },
    "enumString": {
      "type": "string",
      "enum": [
        "a"
      ]
    },
    "formatString": {
      "type": "string",
      "format": "uri"
    },
    "patternString": {
      "type": "string",
      "pattern": "."
    },
    "coordinates": {
      "type": "array",
      "items": {
        "type": [
          "number",
          "array"
        ],
        "minItems": 1
      },
      "minItems": 1
    }
  }
}
'''


def test_command(monkeypatch, tmpdir):
    with open(path('schema-strict.json'), 'rb') as f:
        schema = f.read()

    p = tmpdir.join('schema.json')
    p.write(schema)

    run_command(monkeypatch, main, ['schema-strict', str(p)])

    assert p.read() == expected


def test_command_no_unique_items(monkeypatch, tmpdir):
    with open(path('schema-strict.json'), 'rb') as f:
        schema = f.read()

    p = tmpdir.join('schema.json')
    p.write(schema)

    run_command(monkeypatch, main, ['schema-strict', '--no-unique-items', str(p)])

    assert 'uniqueItems' not in json.loads(p.read())['properties']['array']


@patch('sys.stderr', new_callable=StringIO)
def test_command_check(stderr, monkeypatch):
    with open(path('schema-strict.json'), 'rb') as f:
        expected = f.read()

    run_command(monkeypatch, main, ['schema-strict', '--check', path('schema-strict.json')])

    with open(path('schema-strict.json'), 'rb') as f:
        assert f.read() == expected

    assert stderr.getvalue() == 'ERROR: tests/fixtures/schema-strict.json is missing validation properties\n'



import pytest

from ocdskit.cli.__main__ import main
from tests import assert_command, assert_command_error, path, run_command


@pytest.mark.parametrize('filename,result', [
    ('record-package_minimal.json', 'record package'),
    ('release-package_minimal.json', 'release package'),
    ('record_minimal.json', 'record'),
    ('release_minimal.json', 'release'),
    ('realdata/compiled-release-1.json', 'compiled release'),
    ('realdata/versioned-release-1.json', 'versioned release'),
    ('release-packages.json', 'a JSON array of release packages'),
    ('release-packages.jsonl', 'concatenated JSON, starting with a JSON array of release packages'),
    ('detect-format_mixed.json', 'concatenated JSON, starting with release'),
    ('detect-format_whitespace.json', 'release'),
])
def test_command(filename, result, monkeypatch):
    expected = 'tests/fixtures/{}: {}\n'.format(filename, result)
    assert_command(monkeypatch, main, ['detect-format', path(filename)], expected)


@pytest.mark.parametrize('filename,root_path,result', [
    ('record-package_minimal.json', 'records', 'a JSON array of records'),
    ('record-package_minimal.json', 'records.item', 'record'),
])
def test_command_root_path(filename, root_path, result, monkeypatch):
    expected = 'tests/fixtures/{}: {}\n'.format(filename, result)
    assert_command(monkeypatch, main, ['detect-format', '--root-path', root_path, path(filename)], expected)


def test_command_root_path_nonexistent(monkeypatch, caplog):
    assert_command_error(monkeypatch, main, ['detect-format', '--root-path', 'nonexistent',
                                             path('record_minimal.json')], error=StopIteration)

    assert len(caplog.records) == 0


def test_command_recursive(monkeypatch, caplog, tmpdir):
    content = b'{"records":[]}'
    tmpdir.join('test.json').write(content)
    tmpdir.join('.test.json').write(content)

    actual = run_command(monkeypatch, main, ['detect-format', '--recursive', str(tmpdir)])

    assert '/test.json: record package' in actual
    assert '.test.json' not in actual
    assert len(caplog.records) == 0


@pytest.mark.parametrize('basename,result', [
    ('false', 'boolean'),
    ('null', 'null'),
    ('number', 'number'),
    ('string', 'string'),
    ('true', 'boolean'),
    ('array', 'non-OCDS array'),
    ('object', 'non-OCDS object'),
])
def test_command_unknown_format(basename, result, monkeypatch, caplog):
    filename = 'detect-format_{}.json'.format(basename)

    assert_command(monkeypatch, main, ['detect-format', path(filename)], '')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert caplog.records[0].message == 'tests/fixtures/{}: unknown (top-level JSON value is a {})'.format(
        filename, result)


def test_command_directory(monkeypatch, caplog, tmpdir):
    assert_command(monkeypatch, main, ['detect-format', str(tmpdir)], '')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert caplog.records[0].message.endswith(' is a directory. Set --recursive to recurse into directories.')


def test_command_nonexistent(monkeypatch, caplog):
    assert_command(monkeypatch, main, ['detect-format', 'nonexistent'], '')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert caplog.records[0].message == 'nonexistent: No such file or directory'



from ocdskit.cli.commands.base import OCDSCommand


class Command(OCDSCommand):
    name = 'split-release-packages'
    help = 'reads release packages from standard input, and prints many release packages for each'

    def add_arguments(self):
        self.add_argument('size', type=int, help='the number of releases per package')

    def handle(self):
        # See exploration of not reading each input into memory: https://github.com/open-contracting/ocdskit/issues/118
        for package in self.items():
            releases = package['releases']

            for i in range(0, len(releases), self.args.size):
                package.update(releases=releases[i:i + self.args.size])

                self.print(package)



import json
import logging
import os.path

from ocdskit.cli.commands.base import BaseCommand
from ocdskit.util import json_dump

logger = logging.getLogger('ocdskit')


class Command(BaseCommand):
    name = 'indent'
    help = 'indents JSON files by modifying the given files in-place'

    def add_arguments(self):
        self.add_argument('file', help='files to reindent', nargs='+')
        self.add_argument('-r', '--recursive', help='recursively indent JSON files', action='store_true')
        self.add_argument('--indent', help='indent level', type=int, default=2)

    def handle(self):
        for file in self.args.file:
            if os.path.isfile(file):
                self.indent(file)
            elif self.args.recursive:
                for root, _, files in os.walk(file):
                    for name in files:
                        if name.endswith('.json'):
                            self.indent(os.path.join(root, name))
            elif os.path.isdir(file):
                logger.warning('%s is a directory. Set --recursive to recurse into directories.', file)
            else:
                logger.error('%s: No such file or directory', file)

    def indent(self, path):
        try:
            with open(path) as f:
                data = json.load(f)

            with open(path, 'w') as f:
                json_dump(data, f, indent=self.args.indent, ensure_ascii=self.args.ascii)
                f.write('\n')
        except json.decoder.JSONDecodeError as e:
            logger.error('%s is not valid JSON. (json.decoder.JSONDecodeError: %s)', path, e)
